from enum import Enum


class FieldType(Enum):
    field = "field"
    many_to_many = "many_to_many"
    foreign_key = "foreign_key"
    reverse_foreign_key = "reverse_foreign_key"
    one_to_one = "one_to_one"



import logging


class ExtraFormatter(logging.Formatter):
    def_keys = [
        'name', 'msg', 'args', 'levelname', 'levelno',
        'pathname', 'filename', 'module', 'exc_info',
        'exc_text', 'stack_info', 'lineno', 'funcName',
        'created', 'msecs', 'relativeCreated', 'thread',
        'threadName', 'processName', 'process', 'message'
    ]

    def format(self, record):
        string = super().format(record)
        extra = {k: v for k, v in record.__dict__.items()
                 if k not in self.def_keys}
        if len(extra) > 0:
            string += " - extra: " + str(extra)
        return string



from enum import Enum

from django.core.serializers.json import DjangoJSONEncoder


class EnhancedDjangoJSONEncoder(DjangoJSONEncoder):
    """Default DjangoJSONEncoder does not support Enum"""

    def default(self, o):
        if isinstance(o, Enum):
            # Currently support enum as value
            return o.value

        # Call super to continue with the default encoders
        return super().default(o)



from dataclasses import dataclass

from django.db.models.fields.related import ForeignObjectRel

from fixtures_extractor.enums import FieldType


@dataclass
class ModelFieldMetaDTO:
    app_name: str
    field_name: str
    model_name: str
    field_type: FieldType
    is_model_declared: bool = True

    @staticmethod
    def build(field):
        app_name = field.model._meta.app_label
        model_name = field.model._meta.model_name
        field_name = field.name
        field_type = ModelFieldMetaDTO._detect_field_type(field=field)
        is_model_declared = ModelFieldMetaDTO._detect_model_declaration(field=field)

        if is_model_declared and field_type != FieldType.field:
            app_name = field.related_model._meta.app_label
            model_name = field.related_model._meta.model_name

        if field.is_relation and not is_model_declared:
            app_name = field.related_model._meta.app_label
            model_name = field.related_model._meta.model_name
            field_name = field.field.name

        return ModelFieldMetaDTO(
            app_name=app_name,
            field_name=field_name,
            model_name=model_name,
            field_type=field_type,
            is_model_declared=is_model_declared,
        )

    @classmethod
    def _detect_field_type(cls, field) -> FieldType:
        if not field.is_relation:
            return FieldType.field

        if field.many_to_many:
            return FieldType.many_to_many

        if field.one_to_one:
            return FieldType.one_to_one

        if field.many_to_one:
            return FieldType.foreign_key

        if field.one_to_many:
            return FieldType.reverse_foreign_key

        raise ValueError(f"Unknown field type: {field}")

    @classmethod
    def _detect_model_declaration(cls, field) -> bool:
        return not isinstance(field, ForeignObjectRel)

    def __eq__(self, value: object) -> bool:
        if self.app_name != value.app_name:
            return False

        if self.field_name != value.field_name:
            return False

        if self.model_name != value.model_name:
            return False

        if self.field_type != value.field_type:
            return False

        if self.is_model_declared != value.is_model_declared:
            return False

        return True

    def __gt__(self, value: object) -> bool:
        return self.field_name > value.field_name

    def __lt__(self, value: object) -> bool:
        return self.field_name < value.field_name



import json
import logging
from pathlib import Path
from typing import List

from django.apps import apps

from fixtures_extractor.dtos import ModelFieldMetaDTO
from fixtures_extractor.encoders import EnhancedDjangoJSONEncoder
from fixtures_extractor.enums import FieldType

logger = logging.getLogger()


class ORMExtractor:
    def get_records(self, app_model: str, filter_key: str, filter_value: str):
        fields = self.get_model_fields(app_model=app_model)
        field_names = [field.field_name for field in fields]

        one_relation_fields = self.get_model_declared_one_relations(app_model=app_model)
        one_relation_field_names = [field.field_name for field in one_relation_fields]

        many_to_many_fields = self.get_model_declared_many_relations(
            app_model=app_model
        )
        many_to_many_field_names = [field.field_name for field in many_to_many_fields]

        model = apps.get_model(app_model)
        records = model.objects.all()

        if filter_key:
            records = records.filter(**{filter_key: filter_value}).all()

        base_record_values = list(
            records.values(
                *field_names,
                *one_relation_field_names,
            )
        )

        for index, record in enumerate(records):
            for m2m_field_name in many_to_many_field_names:
                base_record_values[index][m2m_field_name] = list(
                    getattr(record, m2m_field_name).values_list("id", flat=True)
                )

        return base_record_values

    def dump_records(self, records: List, output_file: Path):
        def record_key(record):
            return (record["model"], record["fields"]["id"])

        records = list({record_key(record): record for record in records}.values())

        logger.info(output_file)

        jsonfy_records = json.dumps(records, cls=EnhancedDjangoJSONEncoder, indent=4)
        with open(output_file, "w+") as output:
            output.writelines(jsonfy_records)

    def build_records(self, app_model: str, records: List) -> list:
        results = []
        for item in records:
            values = {}
            for key, value in item.items():
                if key != "pk":
                    values[key] = value

            item_structure = {"model": app_model, "fields": values}
            results.append(item_structure)

        return results

    def get_all_fields(self, app_model):
        model = apps.get_model(app_label=app_model)
        return sorted(
            [ModelFieldMetaDTO.build(field=field) for field in model._meta.get_fields()]
        )

    def get_model_fields(self, app_model) -> List[ModelFieldMetaDTO]:
        return [
            field
            for field in self.get_all_fields(app_model=app_model)
            if field.field_type == FieldType.field
        ]

    def get_model_declared_many_relations(self, app_model) -> List[ModelFieldMetaDTO]:
        return [
            field
            for field in self.get_all_fields(app_model=app_model)
            if field.is_model_declared and field.field_type == FieldType.many_to_many
        ]

    def get_model_declared_one_relations(self, app_model) -> List[ModelFieldMetaDTO]:
        return [
            field
            for field in self.get_all_fields(app_model=app_model)
            if field.is_model_declared
            and field.field_type in [FieldType.one_to_one, FieldType.foreign_key]
        ]

    def get_model_target_relations(self, app_model) -> List[ModelFieldMetaDTO]:
        return [
            field
            for field in self.get_all_fields(app_model=app_model)
            if not field.is_model_declared and field.field_type != FieldType.field
        ]



import logging
from pathlib import Path
from typing import List

from django.core.management.base import BaseCommand

from fixtures_extractor.extra_logging_formatter import ExtraFormatter
from fixtures_extractor.orm_extractor import ORMExtractor

logger = logging.getLogger("extract_fixture")
console = logging.StreamHandler()
console.setLevel(logging.DEBUG)
console.setFormatter(ExtraFormatter("%(name)s - %(levelname)s - %(message)s"))
logger.addHandler(console)


orm_extractor = ORMExtractor()


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "-a",
            "--app",
            type=str,
            help="App of the start model to dump",
        )
        parser.add_argument(
            "-m",
            "--model",
            type=str,
            help="Name of the start model to dump",
        )
        parser.add_argument(
            "primary_ids",
            type=int,
            nargs="+",
            help="Primary ids of the start model to dump",
        )
        parser.add_argument(
            "-d",
            "--output_dir",
            type=str,
            default="fixtures",
            help="Output dir for all the resulted fixtures",
        )

    def handle(self, *args, **options):
        app_name = options.get("app")
        model_name = options.get("model")
        full_model_name = f"{app_name}.{model_name}"

        filter_key = "id"

        primary_ids = options.get("primary_ids")
        output_dir = Path(options.get("output_dir"))

        for primary_id in primary_ids:
            primary_output_dir = output_dir.joinpath(
                f"{model_name.lower()}_{primary_id}"
            )
            primary_output_dir.mkdir(parents=True, exist_ok=True)
            output_file = primary_output_dir.joinpath(f"{full_model_name}.json")

            records = self.process_fields(
                full_model_name=full_model_name,
                filter_key=filter_key,
                filter_value=str(primary_id),
                history=[],
                origin=full_model_name,
            )

            orm_extractor.dump_records(output_file=output_file, records=records)

    def process_fields(
        self,
        full_model_name: str,
        filter_key: str,
        filter_value: str,
        history: List,
        origin: str,
    ) -> list:
        logger.debug((origin, full_model_name, filter_key, filter_value))
        if (origin, full_model_name, filter_key, filter_value) in history:
            logger.debug("----Skipped----")
            return []

        logger.debug("----Processing----")
        history.append((origin, full_model_name, filter_key, filter_value))
        schema_records = []

        base_model_records = orm_extractor.get_records(
            app_model=full_model_name, filter_key=filter_key, filter_value=filter_value
        )
        jsonfy_records = orm_extractor.build_records(
            app_model=full_model_name, records=base_model_records
        )
        schema_records.extend(jsonfy_records)

        one_declared_fields = orm_extractor.get_model_declared_one_relations(
            app_model=full_model_name
        )
        many_declared_fields = orm_extractor.get_model_declared_many_relations(
            app_model=full_model_name
        )
        target_fields = orm_extractor.get_model_target_relations(
            app_model=full_model_name
        )

        for record in base_model_records:
            for one_declared_field in one_declared_fields:
                full_name = (
                    f"{one_declared_field.app_name}.{one_declared_field.model_name}"
                )
                one_declared_records = self.process_fields(
                    full_model_name=full_name,
                    filter_key="id",
                    filter_value=record[one_declared_field.field_name],
                    history=history,
                    origin=full_model_name,
                )
                schema_records.extend(one_declared_records)

            for many_declared_field in many_declared_fields:
                full_name = (
                    f"{many_declared_field.app_name}.{many_declared_field.model_name}"
                )

                for filter_value in record[many_declared_field.field_name]:
                    many_declared_records = self.process_fields(
                        full_model_name=full_name,
                        filter_key="id",
                        filter_value=filter_value,
                        history=history,
                        origin=full_model_name,
                    )
                    schema_records.extend(many_declared_records)

            for target_field in target_fields:
                full_name = f"{target_field.app_name}.{target_field.model_name}"
                target_records = self.process_fields(
                    full_model_name=full_name,
                    filter_key=target_field.field_name,
                    filter_value=record["id"],
                    history=history,
                    origin=full_model_name,
                )
                schema_records.extend(target_records)

        return schema_records
